from collections.abc import Generator
from typing import Any

from sqlalchemy import create_engine, select, text
from sqlalchemy.orm import Session, sessionmaker

from pyvectordb.distance_function import DistanceFunction
from pyvectordb.driver import VectorDB
from pyvectordb.vector import Vector
from pyvectordb.vector_distance import VectorDistance

from .model import VectorORM, get_vector_orm


class PgvectorDB(VectorDB):
    def __init__(
        self,
        host: str,
        port: str,
        user: str,
        password: str,
        db_name: str,
        collection: str,
        distance_function: DistanceFunction | str = DistanceFunction.L2,
    ) -> None:
        super().__init__(host, port)

        self.db_user = user or self.__raise_value_error("db_user")
        self.db_password = password or self.__raise_value_error("db_password")
        self.db_host = host or self.__raise_value_error("db_host")
        self.db_port = port or self.__raise_value_error("db_port")
        self.db_name = db_name or self.__raise_value_error("db_name")
        self.collection = collection or self.__raise_value_error("collection")
        self.distance_function = distance_function or self.__raise_value_error("distance_function")

        self.__engine = None

        self.__init_engine()
        self.conn = next(self.__get_db_session())

        self.__init_collection()
        self.__vector_orm: VectorORM = get_vector_orm(self.collection)

    @staticmethod
    def __raise_value_error(param: str):
        raise ValueError(f"{param} is required")

    def __init_engine(self) -> None:
        if self.__engine is None:
            # pgvector 0.4.x with SQLAlchemy 2.0
            self.__engine = create_engine(
                f"postgresql://{self.db_user}:{self.db_password}@{self.db_host}:{self.db_port}/{self.db_name}",
                echo=False,
                pool_pre_ping=True,
                connect_args={"options": "-c search_path=public"},
            )

    def __get_db_session(self) -> Generator[Session]:
        SessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=self.__engine)

        db = SessionLocal()
        try:
            yield db
        finally:
            db.close()

    def __init_collection(self) -> None:
        query = f"""
CREATE TABLE IF NOT EXISTS {self.collection} (
    id text PRIMARY KEY,
    embedding vector,
    metadata text,
    created_at timestamptz DEFAULT now()
);
"""
        self.conn.execute(text(query))
        self.conn.commit()

    def insert_vector(self, vector: Vector) -> Vector:
        v = self.__vector_orm(
            id=vector.get_id(),
            embedding=vector.embedding,
            metadata_=vector.metadata_to_string(),
        )

        self.conn.add(v)
        self.conn.commit()

    def insert_vectors(self, vectors: list[Vector]) -> None:
        if len(vectors) == 0:
            return

        v_orms = []
        for vector in vectors:
            v_orms.append(
                self.__vector_orm(
                    id=vector.get_id(),
                    embedding=vector.embedding,
                    metadata_=vector.metadata_to_string(),
                )
            )

        self.conn.add_all(v_orms)
        self.conn.commit()

    def read_vector(self, id: str) -> Vector | None:
        v_orm = self.__read_vector_orm(id)

        if v_orm is None:
            return None

        vector = Vector(
            embedding=v_orm.embedding,
            vector_id=v_orm.id,
            metadata=v_orm.metadata_,
        )
        return vector

    def update_vector(self, vector: Vector) -> Vector:
        v = self.__read_vector_orm(vector.id)
        if v is None:
            raise ValueError("vector not found in database")

        v.embedding = vector.embedding
        v.metadata_ = vector.metadata_to_string()

        self.conn.add(v)
        self.conn.commit()

    def update_vectors(self, vectors: list[Vector]) -> None:
        if len(vectors) == 0:
            return

        v_orms = []
        for vector in vectors:
            v = self.__read_vector_orm(vector.id)
            if v is None:
                raise ValueError(f"vector {vector.id} not found in database")

            v.embedding = vector.embedding
            v.metadata_ = vector.metadata_to_string()
            v_orms.append(v)

        self.conn.add_all(v_orms)
        self.conn.commit()

    def delete_vector(self, id: str) -> None:
        v = self.__read_vector_orm(id)
        self.conn.delete(v)
        self.conn.commit()

    def delete_vectors(self, ids: list[str] | list[Vector]) -> None:
        if len(ids) == 0:
            return

        if isinstance(ids[0], Vector):
            [self.delete_vector(v.id) for v in ids]
        else:
            [self.delete_vector(id_) for id_ in ids]

    def get_neighbor_vectors(
        self,
        vector: Vector,
        n: int = 5,
    ) -> list[VectorDistance]:
        vectordistances = []
        distance_func = self.__get_distance_function(self.distance_function)

        q = self.conn.execute(
            select(self.__vector_orm, distance_func(vector.embedding).label("distance"))
            .order_by(distance_func(vector.embedding))
            .limit(n)
        )
        results: tuple[list[VectorORM]] = q.all()

        for r in results:
            vector = Vector(
                embedding=r[0].embedding,
                vector_id=r[0].id,
                metadata=r[0].metadata_,
            )
            distance = r[1]
            vectordistances.append(VectorDistance(vector, distance))
        return vectordistances

    def __get_distance_function(self, distance_function: DistanceFunction | str) -> Any:
        if isinstance(distance_function, str):
            distance_function = DistanceFunction.from_str(distance_function)

        if distance_function == DistanceFunction.L2:
            return self.__vector_orm.embedding.l2_distance
        elif distance_function == DistanceFunction.MAX_INNER_PRODUCT:
            return self.__vector_orm.embedding.max_inner_product
        elif distance_function == DistanceFunction.COSINE:
            return self.__vector_orm.embedding.cosine_distance
        elif distance_function == DistanceFunction.L1:
            return self.__vector_orm.embedding.l1_distance
        elif distance_function == DistanceFunction.HAMMING:
            return self.__vector_orm.embedding.hamming_distance
        elif distance_function == DistanceFunction.JACCARD:
            return self.__vector_orm.embedding.jaccard_distance
        else:
            d_ = [
                "L2",
                "MAX_INNER_PRODUCT",
                "COSINE",
                "L1",
                "HAMMING",
                "JACCARD",
            ]
            raise ValueError(f"distance function unavailable on pgvector: : {d_}")

    def __read_vector_orm(self, id: str) -> VectorORM | None:
        v = self.conn.execute(select(self.__vector_orm).where(self.__vector_orm.id == id)).one_or_none()

        if v is None:
            return None

        return v[0]


__all__ = ["PgvectorDB"]



import logging
import socket
from abc import ABC, abstractmethod

from .vector import Vector
from .vector_distance import VectorDistance


class VectorDB(ABC):
    def __init__(self, host, port, debug: bool = False):
        logging.basicConfig(
            level=logging.DEBUG if debug else logging.INFO,
            format="%(asctime)s - %(name)s - %(levelname)s - %(filename)s - Line: %(lineno)d - %(funcName)s - %(message)s",
            handlers=[
                logging.StreamHandler(),
            ],
        )
        self.__log = logging.getLogger(self.__class__.__name__)

        self.__test_connection(host, port)

    @abstractmethod
    def insert_vector(self, vector: Vector) -> None: ...

    @abstractmethod
    def insert_vectors(self, vectors: list[Vector]) -> None: ...

    @abstractmethod
    def read_vector(self, id: str) -> Vector | None: ...

    @abstractmethod
    def update_vector(self, vector: Vector) -> None: ...

    @abstractmethod
    def update_vectors(self, vectors: list[Vector]) -> None: ...

    @abstractmethod
    def delete_vector(self, id: str) -> None: ...

    @abstractmethod
    def delete_vectors(self, ids: list[str] | list[Vector]) -> None: ...

    @abstractmethod
    def get_neighbor_vectors(self, vector: Vector, n: int) -> list[VectorDistance]: ...

    def __test_connection(self, host, port):
        timeout = 3.0

        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.settimeout(timeout)
            try:
                s.connect((host, int(port)))
                self.__log.info(f"Connection to {host}:{port} succeeded.")

            except (TimeoutError, OSError) as e:
                err_msg = f"Connection to {host}:{port} failed: {e}"
                self.__log.error(err_msg)
                raise ConnectionError(err_msg)



from pinecone import AwsRegion, CloudProvider, Metric, Pinecone, ServerlessSpec

from pyvectordb.distance_function import DistanceFunction
from pyvectordb.driver import VectorDB
from pyvectordb.vector import Vector
from pyvectordb.vector_distance import VectorDistance


class PineconeDB(VectorDB):
    def __init__(
        self,
        api_key: str,
        host: str = None,
        index_name: str = None,
        dimension: int = None,
        cloud: str = "aws",
        region: str = "us-east-1",
        distance_function: DistanceFunction | str = DistanceFunction.COSINE,
        debug: bool = False,
    ) -> None:
        # Pinecone is a managed service, so we use a dummy port for the base class
        super().__init__(host or "pinecone.io", 443, debug)

        self.api_key = api_key or self.__raise_value_error("api_key")
        self.host = host
        self.index_name = index_name
        self.dimension = dimension
        self.cloud = cloud
        self.region = region
        self.distance_function = distance_function or self.__raise_value_error("distance_function")

        self.client: Pinecone = None
        self.index = None

        self.__init_client()
        self.__init_index()

    @staticmethod
    def __raise_value_error(param: str):
        raise ValueError(f"{param} is required")

    def __init_client(self) -> None:
        if self.client is None:
            self.client = Pinecone(api_key=self.api_key)

    def __init_index(self) -> None:
        # If index_name is provided but host is not, try to get host from existing index
        if self.index_name and not self.host:
            try:
                index_description = self.client.describe_index(self.index_name)
                self.host = index_description.host
            except Exception:
                # Index doesn't exist, will create if dimension is provided
                pass

        # Create index if it doesn't exist and dimension is provided
        if self.index_name and self.dimension and not self.host:
            self.__create_index()

        # Initialize index client
        if self.host:
            self.index = self.client.Index(host=self.host)
        elif self.index_name:
            # Try connecting by name (SDK will fetch host automatically)
            self.index = self.client.Index(name=self.index_name)

    def __create_index(self) -> None:
        metric = self.__get_distance_function(self.distance_function)

        self.client.create_index(
            name=self.index_name,
            dimension=self.dimension,
            metric=metric,
            spec=ServerlessSpec(cloud=self.__get_cloud_provider(self.cloud), region=self.__get_aws_region(self.region)),
        )

        # Get the host after creation
        index_description = self.client.describe_index(self.index_name)
        self.host = index_description.host
        self.index = self.client.Index(host=self.host)

    def __get_cloud_provider(self, cloud: str) -> CloudProvider:
        cloud = cloud.lower()
        if cloud == "aws":
            return CloudProvider.AWS
        elif cloud == "gcp":
            return CloudProvider.GCP
        elif cloud == "azure":
            return CloudProvider.AZURE
        else:
            return CloudProvider.AWS

    def __get_aws_region(self, region: str) -> AwsRegion:
        try:
            return AwsRegion(region)
        except ValueError:
            # Return the region string as-is, Pinecone accepts it
            return region

    def __get_distance_function(self, distance_function: DistanceFunction | str) -> Metric:
        if isinstance(distance_function, str):
            distance_function = DistanceFunction.from_str(distance_function)

        if distance_function == DistanceFunction.COSINE:
            return Metric.COSINE
        elif distance_function == DistanceFunction.EUCLIDEAN:
            return Metric.EUCLIDEAN
        elif distance_function == DistanceFunction.DOT:
            return Metric.DOTPRODUCT
        else:
            d_ = [
                "COSINE",
                "EUCLIDEAN",
                "DOT",
            ]
            raise ValueError(f"distance function unavailable on pinecone: {d_}")

    def insert_vector(self, vector: Vector) -> None:
        self.index.upsert(vectors=[(vector.get_id(), vector.embedding, vector.metadata)])

    def insert_vectors(self, vectors: list[Vector]) -> None:
        if len(vectors) == 0:
            return

        vectors_data = [(vector.get_id(), vector.embedding, vector.metadata) for vector in vectors]

        self.index.upsert(vectors=vectors_data)

    def read_vector(self, id: str) -> Vector | None:
        fetch_response = self.index.fetch(ids=[id])

        if id not in fetch_response.vectors:
            return None

        vector_data = fetch_response.vectors[id]
        return Vector(
            embedding=vector_data.values,
            vector_id=vector_data.id,
            metadata=vector_data.metadata,
        )

    def update_vector(self, vector: Vector) -> None:
        # Pinecone uses upsert for both insert and update
        self.insert_vector(vector)

    def update_vectors(self, vectors: list[Vector]) -> None:
        # Pinecone uses upsert for both insert and update
        self.insert_vectors(vectors)

    def delete_vector(self, id: str) -> None:
        self.index.delete(ids=[id])

    def delete_vectors(self, ids: list[str] | list[Vector]) -> None:
        if len(ids) == 0:
            return

        if isinstance(ids[0], Vector):
            ids = [v.get_id() for v in ids]

        self.index.delete(ids=ids)

    def get_neighbor_vectors(
        self,
        vector: Vector,
        n: int,
    ) -> list[VectorDistance]:
        query_response = self.index.query(
            vector=vector.embedding,
            top_k=n,
            include_metadata=True,
            include_values=True,
        )

        vector_distances = []
        for match in query_response.matches:
            vector_distance = VectorDistance(
                vector=Vector(
                    embedding=match.values,
                    vector_id=match.id,
                    metadata=match.metadata,
                ),
                distance=match.score,
            )
            vector_distances.append(vector_distance)

        return vector_distances


__all__ = ["PineconeDB"]



from enum import Enum


class DistanceFunction(Enum):
    """Distance function to measure how var the vector to surrounding vector in vector db"""

    COSINE = "cosine"
    HAMMING = "hamming"
    JACCARD = "jaccard"
    L1 = "l1"
    L2 = "l2"
    MAX_INNER_PRODUCT = "max_inner_product"
    EUCLIDEAN = "euclid"
    DOT = "dot"
    MANHATTAN = "manhattan"

    @staticmethod
    def from_str(text: str) -> "DistanceFunction":
        text = text.lower()

        if text == "cosine":
            return DistanceFunction.COSINE
        if text == "hamming":
            return DistanceFunction.HAMMING
        if text == "jaccard":
            return DistanceFunction.JACCARD
        if text == "l1":
            return DistanceFunction.L1
        if text == "l2":
            return DistanceFunction.L2
        if text == "max_inner_product":
            return DistanceFunction.MAX_INNER_PRODUCT
        if text == "euclid":
            return DistanceFunction.EUCLIDEAN
        if text == "euclidean":
            return DistanceFunction.EUCLIDEAN
        if text == "dot":
            return DistanceFunction.DOT
        if text == "manhattan":
            return DistanceFunction.MANHATTAN

        raise ValueError("invalid string for distance function")



from enum import Enum


class Distance(Enum):
    """Weaviate distance metrics"""

    COSINE = "cosine"
    DOT = "dot"
    L2_SQUARED = "l2-squared"
    HAMMING = "hamming"
    MANHATTAN = "manhattan"



import chromadb
from chromadb.config import Settings

from pyvectordb.distance_function import DistanceFunction
from pyvectordb.driver import VectorDB
from pyvectordb.vector import Vector
from pyvectordb.vector_distance import VectorDistance

from .distance import Distance


class ChromaDB(VectorDB):
    def __init__(
        self,
        host: str,
        port: str,
        auth_provider: str = None,
        auth_credentials: str = None,
        collection_name: str = None,
        distance_function: DistanceFunction | str = DistanceFunction.L2,
        debug: bool = False,
    ) -> None:
        super().__init__(host, port, debug)

        self.host = host or self.__raise_value_error("host")
        self.port = port or self.__raise_value_error("port")
        self.auth_provider = auth_provider
        self.auth_credentials = auth_credentials
        self.collection_name = collection_name or self.__raise_value_error("collection_name")
        self.distance_function = distance_function or self.__raise_value_error("distance_function")

        self.client = None
        self.collection = None

        self.__init_client()
        self.__health_check()
        self.__init_collection()

    @staticmethod
    def __raise_value_error(param: str):
        raise ValueError(f"{param} is required")

    def __init_client(self) -> None:
        if self.client is None:
            # ChromaDB v1.x API - using HttpClient with settings
            settings = Settings(
                anonymized_telemetry=False,
                allow_reset=True,
            )

            if self.auth_provider and self.auth_credentials:
                settings.chroma_client_auth_provider = self.auth_provider
                settings.chroma_client_auth_credentials = self.auth_credentials

            self.client = chromadb.HttpClient(
                host=self.host,
                port=self.port,
                settings=settings,
            )

    def __health_check(self) -> None:
        self.client.heartbeat()

    def __init_collection(self) -> None:
        if self.collection is None:
            distance_func = self.__get_distance_function(self.distance_function)
            self.collection = self.client.get_or_create_collection(
                name=self.collection_name, metadata={"hnsw:space": distance_func.value}
            )

    def insert_vector(self, vector: Vector) -> None:
        self.collection.add(
            ids=[vector.get_id()],
            embeddings=[vector.embedding],
            metadatas=[vector.metadata],
        )

    def insert_vectors(self, vectors: list[Vector]) -> None:
        if len(vectors) == 0:
            return

        ids, embeddings, metadatas = [], [], []
        for v in vectors:
            ids.append(v.get_id())
            embeddings.append(v.embedding)
            metadatas.append(v.metadata)

        self.collection.add(
            ids=ids,
            embeddings=embeddings,
            metadatas=metadatas,
        )

    def read_vector(self, id: str) -> Vector | None:
        result: dict = self.collection.get(ids=id, include=["metadatas", "documents", "embeddings"])

        if len(result.get("ids")) == 0:
            return None

        vector_id = result.get("ids")[0]
        embedding = result.get("embeddings")[0]
        metadata = result.get("metadatas")[0]

        return Vector(
            embedding=embedding,
            vector_id=vector_id,
            metadata=metadata,
        )

    def update_vector(self, vector: Vector) -> None:
        self.collection.update(
            ids=[vector.get_id()],
            embeddings=[vector.embedding],
            metadatas=[vector.metadata],
        )

    def update_vectors(self, vectors: list[Vector]) -> None:
        if len(vectors) == 0:
            return

        ids, embeddings, metadatas = [], [], []
        for v in vectors:
            ids.append(v.id)
            embeddings.append(v.embedding)
            metadatas.append(v.metadata)

        self.collection.update(
            ids=ids,
            embeddings=embeddings,
            metadatas=metadatas,
        )

    def delete_vector(self, id_: str) -> None:
        self.collection.delete(
            ids=[id_],
        )

    def delete_vectors(self, ids: list[str] | list[Vector]) -> None:
        if len(ids) == 0:
            return

        if isinstance(ids[0], Vector):
            ids = [v.id for v in ids]

        self.collection.delete(
            ids=ids,
        )

    def get_neighbor_vectors(
        self,
        vector: Vector,
        n: int,
    ) -> list[VectorDistance]:
        result: dict = self.collection.query(
            query_embeddings=[vector.embedding],
            n_results=n,
            include=["metadatas", "documents", "distances", "embeddings"],
        )

        vds = []
        for i, _ in enumerate(result.get("ids")[0]):
            vd = VectorDistance(
                vector=Vector(
                    vector_id=result.get("ids")[0][i],
                    embedding=result.get("embeddings")[0][i],
                    metadata=result.get("metadatas")[0][i],
                ),
                distance=result.get("distances")[0][i],
            )
            vds.append(vd)

        return vds

    def __get_distance_function(self, distance_function: DistanceFunction | str) -> Distance:
        if isinstance(distance_function, str):
            distance_function = DistanceFunction.from_str(distance_function)

        if distance_function == DistanceFunction.L2:
            return Distance.SQUARED_L2
        elif distance_function == DistanceFunction.MAX_INNER_PRODUCT:
            return Distance.INNER_PRODUCT
        elif distance_function == DistanceFunction.COSINE:
            return Distance.COSINE_SIMILARITY
        else:
            d_ = [
                "L2",
                "MAX_INNER_PRODUCT",
                "COSINE",
            ]
            raise ValueError(f"distance function unavailable on chromadb. {d_}")


__all__ = ["ChromaDB"]



from datetime import datetime

from pgvector.sqlalchemy import Vector
from sqlalchemy import Column, DateTime, String
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column


class Base(DeclarativeBase):
    __abstract__ = True

    def to_dict(self):
        return {field.name: getattr(self, field.name) for field in self.__table__.c}


class VectorORM(Base):
    __abstract__ = True

    id: Mapped[str] = Column(String, primary_key=True)
    embedding: Mapped[Vector] = mapped_column(Vector())
    metadata_: Mapped[str] = Column(String, nullable=False, name="metadata")
    created_at: Mapped[datetime] = Column(DateTime, nullable=False, default=datetime.now())


def get_vector_orm(tablename: str) -> "VectorORM":
    class VectorORMreal(VectorORM):
        __tablename__ = tablename

    return VectorORMreal



import json
from uuid import uuid4


class Vector:
    def __init__(
        self,
        embedding: list[float],
        vector_id: str | None = None,
        metadata: dict | str | None = None,
        init_id: bool = False,
    ) -> None:

        if embedding is None or len(embedding) == 0:
            self.__raise_value_error("embedding")
        self.embedding = embedding if embedding is not None else {}

        self.id = vector_id
        self.metadata = self.metadata_from_string(metadata) if isinstance(metadata, str) else metadata

        if init_id:
            self.get_id()

    @staticmethod
    def __raise_value_error(param: str):
        raise ValueError(f"{param} is required")

    def get_id(self) -> str:
        if self.id is None:
            self.id = str(uuid4())
        return self.id

    def metadata_to_string(self) -> str:
        return json.dumps(self.metadata)

    def metadata_from_string(self, metadata: str) -> dict:
        self.metadata = json.loads(metadata)
        return self.metadata

    def __len__(self) -> int:
        return len(self.embedding)

    def __str__(self) -> str:
        if self.embedding is not None and len(self.embedding) > 10:
            embedding = f"{self.embedding[:5]}...{self.embedding[-5:]}"
        else:
            embedding = self.embedding

        if self.metadata and len(self.metadata) > 10:
            metadata = f"{self.metadata[:5]}...{self.metadata[-5:]}"
        else:
            metadata = self.metadata

        return f"""Vector[id: {self.id}, embedding: {embedding}, embedding_length: {len(self.embedding)}, metadata: {metadata}]"""

    def __repr__(self) -> str:
        return self.__str__()



from pymilvus import MilvusClient

from pyvectordb.distance_function import DistanceFunction
from pyvectordb.driver import VectorDB
from pyvectordb.vector import Vector
from pyvectordb.vector_distance import VectorDistance

from .distance import Distance


class MilvusDB(VectorDB):
    def __init__(
        self,
        host: str,
        port: int = 19530,
        collection: str = None,
        vector_size: int = None,
        distance_function: DistanceFunction | str = DistanceFunction.COSINE,
        debug: bool = False,
    ) -> None:
        super().__init__(host, port, debug)

        self.host = host or self.__raise_value_error("host")
        self.port = port or self.__raise_value_error("port")
        self.collection = collection or self.__raise_value_error("collection")
        self.vector_size = vector_size or self.__raise_value_error("vector_size")
        self.distance_function = distance_function or self.__raise_value_error("distance_function")

        self.client: MilvusClient = None

        self.__init_client()
        self.__init_collection()

    @staticmethod
    def __raise_value_error(param: str):
        raise ValueError(f"{param} is required")

    def __init_client(self) -> None:
        if self.client is None:
            uri = f"http://{self.host}:{self.port}"
            self.client = MilvusClient(uri=uri)

    def __init_collection(self) -> None:
        if not self.client.has_collection(self.collection):
            metric_type = self.__get_distance_function(self.distance_function)
            self.client.create_collection(
                collection_name=self.collection,
                dimension=self.vector_size,
                metric_type=metric_type,
            )

    def __get_distance_function(self, distance_function: DistanceFunction | str) -> str:
        if isinstance(distance_function, str):
            distance_function = DistanceFunction.from_str(distance_function)

        if distance_function == DistanceFunction.COSINE:
            return Distance.COSINE.value
        elif distance_function == DistanceFunction.EUCLIDEAN:
            return Distance.L2.value
        elif distance_function == DistanceFunction.DOT:
            return Distance.IP.value
        elif distance_function == DistanceFunction.HAMMING:
            return Distance.HAMMING.value
        elif distance_function == DistanceFunction.JACCARD:
            return Distance.JACCARD.value
        else:
            d_ = [
                "COSINE",
                "EUCLIDEAN",
                "DOT",
                "HAMMING",
                "JACCARD",
            ]
            raise ValueError(f"distance function unavailable on milvus: {d_}")

    def insert_vector(self, vector: Vector) -> None:
        vector_id = vector.get_id()

        self.client.insert(
            collection_name=self.collection,
            data=[
                {
                    "id": vector_id,
                    "vector": vector.embedding,
                    "metadata": vector.metadata,
                }
            ],
        )

    def insert_vectors(self, vectors: list[Vector]) -> None:
        if len(vectors) == 0:
            return

        data = []
        for vector in vectors:
            data.append(
                {
                    "id": vector.get_id(),
                    "vector": vector.embedding,
                    "metadata": vector.metadata,
                }
            )

        self.client.insert(collection_name=self.collection, data=data)

    def read_vector(self, id: str) -> Vector | None:
        results = self.client.query(
            collection_name=self.collection,
            ids=[id],
            output_fields=["vector", "metadata"],
        )

        if len(results) == 0:
            return None

        result = results[0]
        return Vector(
            embedding=result.get("vector"),
            vector_id=result.get("id"),
            metadata=result.get("metadata"),
        )

    def update_vector(self, vector: Vector) -> None:
        # Milvus uses upsert for both insert and update
        self.client.upsert(
            collection_name=self.collection,
            data=[
                {
                    "id": vector.get_id(),
                    "vector": vector.embedding,
                    "metadata": vector.metadata,
                }
            ],
        )

    def update_vectors(self, vectors: list[Vector]) -> None:
        if len(vectors) == 0:
            return

        data = []
        for vector in vectors:
            data.append(
                {
                    "id": vector.get_id(),
                    "vector": vector.embedding,
                    "metadata": vector.metadata,
                }
            )

        self.client.upsert(collection_name=self.collection, data=data)

    def delete_vector(self, id: str) -> None:
        self.client.delete(
            collection_name=self.collection,
            ids=[id],
        )

    def delete_vectors(self, ids: list[str] | list[Vector]) -> None:
        if len(ids) == 0:
            return

        if isinstance(ids[0], Vector):
            ids = [v.get_id() for v in ids]

        self.client.delete(
            collection_name=self.collection,
            ids=ids,
        )

    def get_neighbor_vectors(
        self,
        vector: Vector,
        n: int,
    ) -> list[VectorDistance]:
        results = self.client.search(
            collection_name=self.collection,
            data=[vector.embedding],
            limit=n,
            output_fields=["id", "vector", "metadata"],
        )

        vector_distances = []
        for hit in results[0]:
            vector_distance = VectorDistance(
                vector=Vector(
                    embedding=hit.get("entity", {}).get("vector"),
                    vector_id=hit.get("id"),
                    metadata=hit.get("entity", {}).get("metadata"),
                ),
                distance=hit.get("distance", 0.0),
            )
            vector_distances.append(vector_distance)

        return vector_distances


__all__ = ["MilvusDB"]



import os

from dotenv import load_dotenv

from pyvectordb import Vector
from pyvectordb.distance_function import DistanceFunction
from pyvectordb.pgvector import PgvectorDB

load_dotenv()

v1 = Vector(embedding=[2.0, 2.0, 1.0], metadata={"text": "hellow from pyvectordb"})
v2 = Vector(embedding=[2.0, 2.0, 2.0], metadata={"text": "hi"})
v3 = Vector(embedding=[2.0, 2.0, 3.0], metadata={"text": "good morning!"})

vector_db = PgvectorDB(
    user=os.getenv("PG_USER"),
    password=os.getenv("PG_PASSWORD"),
    host=os.getenv("PG_HOST"),
    port=os.getenv("PG_PORT"),
    db_name=os.getenv("PG_NAME"),
    collection=os.getenv("PG_COLLECTION"),
    distance_function=DistanceFunction.L2,
)

# insert new vector
vector_db.insert_vector(v1)
vector_db.insert_vectors([v2, v3])

# read v1
v_from_db = vector_db.read_vector(v1.get_id())

# update v1 embedding
new_embedding = [2.0, 2.0, 4.0]
v_from_db.embedding = new_embedding
vector_db.update_vector(v_from_db)

# read updated embedding and check
v_from_db_updated = vector_db.read_vector(v1.get_id())
assert list(v_from_db_updated.embedding) == list(new_embedding), "updated embedding not equal"

# re-update v1 embedding to the v1, check
vector_db.update_vectors([v1, v2, v3])
re_updated_embedding = vector_db.read_vector(v1.get_id()).embedding
assert list(re_updated_embedding) == list(v1.embedding), "re-updated embedding not equal"

for x in vector_db.get_neighbor_vectors(v1, 3):
    print(f"{x}")

vector_db.delete_vector(v1.get_id())
vector_db.delete_vectors([v2, v3])



from qdrant_client import QdrantClient
from qdrant_client.models import Distance, PointStruct, ScoredPoint, VectorParams

from pyvectordb.distance_function import DistanceFunction
from pyvectordb.driver import VectorDB
from pyvectordb.vector import Vector
from pyvectordb.vector_distance import VectorDistance


class QdrantDB(VectorDB):
    def __init__(
        self,
        host: str,
        api_key: str = None,
        port: int = 6333,
        collection: str = None,
        vector_size: int = None,
        distance_function: DistanceFunction | str = DistanceFunction.EUCLIDEAN,
    ) -> None:
        super().__init__(host, port)

        self.host = host or self.__raise_value_error("host")
        self.api_key = api_key
        self.port = port or self.__raise_value_error("port")
        self.collection = collection or self.__raise_value_error("collection")
        self.vector_size = vector_size or self.__raise_value_error("vector_size")
        self.distance_function = distance_function or self.__raise_value_error("distance_function")

        self.client: QdrantClient = None

        self.__init__client()
        self.__init_collection()

    @staticmethod
    def __raise_value_error(param: str):
        raise ValueError(f"{param} is required")

    def __init__client(self) -> None:
        if self.client is None:
            # Qdrant client v1.16.x API
            self.client = QdrantClient(
                host=self.host,
                port=self.port,
                api_key=self.api_key,
                https=False,
                timeout=10,
            )

    def __init_collection(self) -> None:
        if not self.client.collection_exists(self.collection):
            self.client.create_collection(
                collection_name=self.collection,
                vectors_config=VectorParams(
                    size=self.vector_size, distance=self.__get_distance_function(self.distance_function)
                ),
            )

    def __get_distance_function(self, distance_function: DistanceFunction | str) -> Distance:
        if isinstance(distance_function, str):
            distance_function = DistanceFunction.from_str(distance_function)

        if distance_function == DistanceFunction.COSINE:
            distance_func = Distance.COSINE
        elif distance_function == DistanceFunction.EUCLIDEAN:
            distance_func = Distance.EUCLID
        elif distance_function == DistanceFunction.DOT:
            distance_func = Distance.DOT
        elif distance_function == DistanceFunction.MANHATTAN:
            distance_func = Distance.MANHATTAN
        else:
            d_ = [
                "COSINE",
                "EUCLIDEAN",
                "DOT",
                "MANHATTAN",
            ]
            raise ValueError(f"distance function unavailable on qdrant: {d_}")

        return distance_func

    def insert_vector(self, vector: Vector) -> None:
        vector_id = vector.get_id()

        self.client.upsert(
            collection_name=self.collection,
            points=[PointStruct(id=vector_id, vector=vector.embedding, payload={"metadata": vector.metadata})],
            wait=True,
        )

    def insert_vectors(self, vectors: list[Vector]) -> None:
        if len(vectors) == 0:
            return

        points = [
            PointStruct(
                id=vector.get_id(),
                vector=vector.embedding,
                payload=vector.metadata,
            )
            for vector in vectors
        ]

        self.client.upsert(collection_name=self.collection, points=points, wait=False)

    def read_vector(self, id: str) -> Vector | None:
        records = self.client.retrieve(
            collection_name=self.collection,
            ids=[id],
            with_payload=True,
            with_vectors=True,
        )
        if len(records) == 0:
            return None

        record = records[0]
        return Vector(
            embedding=record.vector,
            vector_id=record.id,
            metadata=record.payload.get("metadata"),
        )

    def update_vector(self, vector: Vector) -> None:
        # we use qdrant upsert, so...
        self.insert_vector(vector)

    def update_vectors(self, vectors: list[Vector]) -> None:
        # we use qdrant upsert, so...
        self.insert_vectors(vectors)

    def delete_vector(self, id: str) -> None:
        self.client.delete(collection_name=self.collection, points_selector=[id], wait=False)

    def delete_vectors(self, ids: list[str] | list[Vector]) -> None:
        if len(ids) == 0:
            return

        if isinstance(ids[0], Vector):
            [self.delete_vector(v.id) for v in ids]
        else:
            [self.delete_vector(id_) for id_ in ids]

    def get_neighbor_vectors(
        self,
        vector: Vector,
        n: int,
    ) -> list[VectorDistance]:
        scored_points: list[ScoredPoint] = self.client.search(
            collection_name=self.collection,
            query_vector=vector.embedding,
            with_payload=True,
            with_vectors=True,
            limit=n,
        )
        vector_distances = []
        for point in scored_points:
            vector_distance = VectorDistance(
                vector=Vector(
                    embedding=point.vector,
                    vector_id=point.id,
                    metadata=point.payload.get("metadata"),
                ),
                distance=point.score,
            )
            vector_distances.append(vector_distance)
        return vector_distances


__all__ = ["QdrantDB"]



import os

from dotenv import load_dotenv

from pyvectordb import Vector
from pyvectordb.distance_function import DistanceFunction
from pyvectordb.weaviate import WeaviateDB

load_dotenv()

v1 = Vector(embedding=[2.0, 2.0, 1.0], metadata={"text": "hello from pyvectordb"})
v2 = Vector(embedding=[2.0, 2.0, 2.0], metadata={"text": "hi"})
v3 = Vector(embedding=[2.0, 2.0, 3.0], metadata={"text": "good morning!"})

# Initialize WeaviateDB
# Option 1: Connect to local Weaviate instance
vector_db = WeaviateDB(
    host=os.getenv("WEAVIATE_HOST", "localhost"),
    port=int(os.getenv("WEAVIATE_PORT", 8080)),
    grpc_port=int(os.getenv("WEAVIATE_GRPC_PORT", 50051)),
    api_key=os.getenv("WEAVIATE_API_KEY"),
    collection=os.getenv("WEAVIATE_COLLECTION"),
    vector_size=int(os.getenv("WEAVIATE_VECTOR_SIZE")),
    distance_function=DistanceFunction.COSINE,
)

# insert new vector
vector_db.insert_vector(v1)
vector_db.insert_vectors([v2, v3])

# read v1
v_from_db = vector_db.read_vector(v1.get_id())

# update v1 embedding
new_embedding = [2.0, 2.0, 4.0]
v_from_db.embedding = new_embedding
vector_db.update_vector(v_from_db)

# read updated embedding and check
v_from_db_updated = vector_db.read_vector(v1.get_id())
assert list(v_from_db_updated.embedding) == list(new_embedding), "updated embedding not equal"

# re-update v1 embedding to the v1, check
vector_db.update_vectors([v1, v2, v3])
re_updated_embedding = vector_db.read_vector(v1.get_id()).embedding
assert list(re_updated_embedding) == list(v1.embedding), "re-updated embedding not equal"

for x in vector_db.get_neighbor_vectors(v1, 3):
    print(f"{x}")

vector_db.delete_vector(v1.get_id())
vector_db.delete_vectors([v2, v3])



import weaviate
import weaviate.classes.config as wvc
from weaviate.classes.query import MetadataQuery

from pyvectordb.distance_function import DistanceFunction
from pyvectordb.driver import VectorDB
from pyvectordb.vector import Vector
from pyvectordb.vector_distance import VectorDistance

from .distance import Distance


class WeaviateDB(VectorDB):
    def __init__(
        self,
        host: str = "localhost",
        port: int = 8080,
        grpc_port: int = 50051,
        api_key: str = None,
        collection: str = None,
        vector_size: int = None,
        distance_function: DistanceFunction | str = DistanceFunction.COSINE,
        debug: bool = False,
    ) -> None:
        super().__init__(host, port, debug)

        self.host = host or self.__raise_value_error("host")
        self.port = port or self.__raise_value_error("port")
        self.grpc_port = grpc_port
        self.api_key = api_key
        self.collection_name = collection or self.__raise_value_error("collection")
        self.vector_size = vector_size or self.__raise_value_error("vector_size")
        self.distance_function = distance_function or self.__raise_value_error("distance_function")

        self.client = None
        self.collection = None

        self.__init_client()
        self.__init_collection()

    @staticmethod
    def __raise_value_error(param: str):
        raise ValueError(f"{param} is required")

    def __init_client(self) -> None:
        if self.client is None:
            if self.api_key:
                # Connect to Weaviate Cloud
                self.client = weaviate.connect_to_weaviate_cloud(
                    cluster_url=self.host,
                    auth_credentials=weaviate.classes.init.Auth.api_key(self.api_key),
                )
            else:
                # Connect to local/custom Weaviate instance
                self.client = weaviate.connect_to_custom(
                    http_host=self.host,
                    http_port=self.port,
                    http_secure=False,
                    grpc_host=self.host,
                    grpc_port=self.grpc_port,
                    grpc_secure=False,
                )

    def __init_collection(self) -> None:
        if self.collection is None:
            # Check if collection exists
            try:
                self.collection = self.client.collections.get(self.collection_name)
                # Verify the collection exists by trying to fetch
                self.collection.collections.get()
            except Exception:
                # Create collection if it doesn't exist
                distance_metric = self.__get_distance_function(self.distance_function)
                self.client.collections.create(
                    name=self.collection_name,
                    vectorizer_config=wvc.Configure.Vectorizer.none(),
                    vector_index_config=wvc.Configure.VectorIndex.flat(distance_metric=distance_metric),
                    properties=[wvc.Property(name="metadata", data_type=wvc.DataType.OBJECT)],
                )
                self.collection = self.client.collections.get(self.collection_name)

    def __get_distance_function(self, distance_function: DistanceFunction | str) -> str:
        if isinstance(distance_function, str):
            distance_function = DistanceFunction.from_str(distance_function)

        if distance_function == DistanceFunction.COSINE:
            return Distance.COSINE.value
        elif distance_function == DistanceFunction.EUCLIDEAN:
            return Distance.L2_SQUARED.value
        elif distance_function == DistanceFunction.DOT:
            return Distance.DOT.value
        elif distance_function == DistanceFunction.MANHATTAN:
            return Distance.MANHATTAN.value
        elif distance_function == DistanceFunction.HAMMING:
            return Distance.HAMMING.value
        else:
            d_ = [
                "COSINE",
                "EUCLIDEAN",
                "DOT",
                "MANHATTAN",
                "HAMMING",
            ]
            raise ValueError(f"distance function unavailable on weaviate: {d_}")

    def insert_vector(self, vector: Vector) -> None:
        vector_id = vector.get_id()

        self.collection.data.insert(
            uuid=vector_id,
            properties={"metadata": vector.metadata},
            vector=vector.embedding,
        )

    def insert_vectors(self, vectors: list[Vector]) -> None:
        if len(vectors) == 0:
            return

        data = []
        for vector in vectors:
            data.append(
                {
                    "uuid": vector.get_id(),
                    "properties": {"metadata": vector.metadata},
                    "vector": vector.embedding,
                }
            )

        # Use insert_many for batch insert
        objects = []
        for item in data:
            from weaviate.classes.data import DataObject

            obj = DataObject(
                uuid=item["uuid"],
                properties=item["properties"],
                vector=item["vector"],
            )
            objects.append(obj)

        self.collection.data.insert_many(objects)

    def read_vector(self, id: str) -> Vector | None:
        try:
            result = self.collection.query.fetch_object_by_id(
                uuid=id,
                include_vector=True,
            )

            if result is None:
                return None

            return Vector(
                embedding=result.vector,
                vector_id=result.uuid,
                metadata=result.properties.get("metadata") if result.properties else None,
            )
        except Exception:
            return None

    def update_vector(self, vector: Vector) -> None:
        vector_id = vector.get_id()

        self.collection.data.update(
            uuid=vector_id,
            properties={"metadata": vector.metadata},
            vector=vector.embedding,
        )

    def update_vectors(self, vectors: list[Vector]) -> None:
        # Weaviate doesn't have a direct batch update, so we update one by one
        for vector in vectors:
            self.update_vector(vector)

    def delete_vector(self, id: str) -> None:
        self.collection.data.delete_by_id(uuid=id)

    def delete_vectors(self, ids: list[str] | list[Vector]) -> None:
        if len(ids) == 0:
            return

        if isinstance(ids[0], Vector):
            ids = [v.get_id() for v in ids]

        for id_ in ids:
            self.delete_vector(id_)

    def get_neighbor_vectors(
        self,
        vector: Vector,
        n: int,
    ) -> list[VectorDistance]:
        results = self.collection.query.near_vector(
            near_vector=vector.embedding,
            limit=n,
            return_metadata=MetadataQuery(distance=True),
        )

        vector_distances = []
        for obj in results.objects:
            vector_distance = VectorDistance(
                vector=Vector(
                    embedding=obj.vector,
                    vector_id=obj.uuid,
                    metadata=obj.properties.get("metadata") if obj.properties else None,
                ),
                distance=obj.metadata.distance if obj.metadata else 0.0,
            )
            vector_distances.append(vector_distance)

        return vector_distances


__all__ = ["WeaviateDB"]



import os

from dotenv import load_dotenv

from pyvectordb import Vector
from pyvectordb.distance_function import DistanceFunction
from pyvectordb.weaviate import WeaviateDB

load_dotenv()


def test_integration():
    v1 = Vector(embedding=[2.0, 2.0, 1.0], metadata={"text": "hello from pyvectordb"})
    v2 = Vector(embedding=[2.0, 2.0, 2.0], metadata={"text": "hi"})
    v3 = Vector(embedding=[2.0, 2.0, 3.0], metadata={"text": "good morning!"})

    vector_db = WeaviateDB(
        host=os.getenv("WEAVIATE_HOST", "localhost"),
        port=int(os.getenv("WEAVIATE_PORT", 8080)),
        grpc_port=int(os.getenv("WEAVIATE_GRPC_PORT", 50051)),
        api_key=os.getenv("WEAVIATE_API_KEY"),
        collection=os.getenv("WEAVIATE_COLLECTION"),
        vector_size=int(os.getenv("WEAVIATE_VECTOR_SIZE")),
        distance_function=DistanceFunction.COSINE,
    )

    # insert new vector
    vector_db.insert_vector(v1)
    vector_db.insert_vectors([v2, v3])

    # read v1
    v_from_db = vector_db.read_vector(v1.get_id())

    # update v1 embedding
    new_embedding = [2.0, 2.0, 4.0]
    v_from_db.embedding = new_embedding
    vector_db.update_vector(v_from_db)

    # read updated embedding and check
    v_from_db_updated = vector_db.read_vector(v1.get_id())
    assert list(v_from_db_updated.embedding) == list(new_embedding), "updated embedding not equal"

    # re-update v1 embedding to the v1, check
    vector_db.update_vectors([v1, v2, v3])
    re_updated_embedding = vector_db.read_vector(v1.get_id()).embedding
    assert list(re_updated_embedding) == list(v1.embedding), "re-updated embedding not equal"

    for x in vector_db.get_neighbor_vectors(v1, 3):
        print(f"{x}")

    vector_db.delete_vector(v1.get_id())
    vector_db.delete_vectors([v2, v3])
